package main

import "fmt"
import "log"
import "net"
import "os"
import "strconv"
import "syscall"
import "unsafe"
import netstructs "./network_structs"
import "./minefield"

var rows, cols, mines, player_id int
var next_player_to_move int
var server_addr string
var server_conn net.Conn
var my_score, enemy_score int
var field *minefield.Minefield
const abc string = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"

func main() {
	init_terminal()
	defer func() {
		restore_terminal()
		if recover() != nil {
			fmt.Println("Exited.")
			os.Exit(1)
		}
	}()
	server_addr = "localhost"
	handle_cmd_arguments()
	connect_to_server()
	handle_game_desc()

	game_loop()
}

func game_loop() {
	for {
		get_field()
		print_field()
		if next_player_to_move == player_id {
			move()
		} else if next_player_to_move < 0 {
			fmt.Print("Game over: ")
			if next_player_to_move+2 == player_id {
				fmt.Println("You've won!")
			} else {
				fmt.Println("You've lost!")
			}
			break
		} else {
			fmt.Println("Waiting for the other player.")
		}
	}
}

func move() {
	fmt.Println("Click on the field to reveal it!")
	var esc, bracket, M, mode byte
	var x, y int
	init_mouse()
	for {
		fmt.Scanf("%c", &esc); if esc != 27 { continue; }
		fmt.Scanf("%c", &bracket); if bracket != '[' { continue; }
		fmt.Scanf("%c", &M); if M != 'M' { continue; }
		fmt.Scanf("%c", &mode); if mode != '#' { continue; }
		fmt.Scanf("%c", &x)
		fmt.Scanf("%c", &y)
		// -32 because coordinates are sent back via printable characters
		// - 3 because the grid is not exactly at the top left corner
		x -= 35
		y -= 35
		fmt.Printf("(%d, %d)\n", x, y)
		if x >= 0 && x < rows || y >= 0 || y < cols {
			break
		}
	}
	restore_mouse()
	var move netstructs.Move
	move.Row = int32(y)
	move.Col = int32(x)
	netstructs.SendMove(move, server_conn)
}

func get_field() {
	var netfield netstructs.Field
	netfield.Fields = make([]byte, rows*cols)
	netstructs.RecvField(&netfield, server_conn)
	next_player_to_move = int(netfield.Next_player)
	my_score = int(netfield.Scores[player_id])
	enemy_score = int(netfield.Scores[1-player_id])
	minefield.FromByteSlice(netfield.Fields, field)
}

func clear_screen() {
	fmt.Printf("\x1B[H\x1B[2J")
}

func print_field() {
	clear_screen()
	fmt.Printf("Mines found: me: %2d, enemy: %2d\n\n", my_score, enemy_score)

	for r := 0; r < rows; r++ {
		fmt.Printf("  ")
		for c := 0; c < cols; c++ {
			var b byte = field.Fields[r][c]
			if b < 9 {
				fmt.Printf("%c", int(b) + '0')
			} else if b == '?' {
				fmt.Print(".")
			} else if b == 'x' {
				fmt.Print("X")
			} else {
				panic("Undefined field value!")
			}
		}
		fmt.Println()
	}
	fmt.Println("\n")
}

func connect_to_server() {
	remote_addr := server_addr + ":" + strconv.Itoa(netstructs.Port)
	log.Printf("Connecting to %s\n", remote_addr)
	var err os.Error
	server_conn, err = net.Dial("tcp", "", remote_addr)
	check_error(err)
}

func handle_game_desc() {
	desc := netstructs.RecvGameDescription(server_conn)
	rows = int(desc.Rows)
	cols = int(desc.Cols)
	mines = int(desc.Mines)
	player_id = int(desc.Player_id)
	log.Printf("Player id: %d\n", player_id)
	log.Printf("Game field is %dx%d with %d mines.\n", rows, cols, mines)
	field = minefield.Create(int32(rows), int32(cols), int32(mines))
}

func handle_cmd_arguments() {
	if len(os.Args) >= 2 && (os.Args[1] == "-h" || os.Args[1] == "--help") {
		fmt.Println("usage  : ", os.Args[0], " [hostname]")
		fmt.Printf("default: %s %s\n", os.Args[0], server_addr)
		restore_terminal()
		os.Exit(0)
	}
}

const tcgets = 0x5401
const tcsetsf = 0x5404
type termios struct {
	c_iflag  uint "input mode flags"
	c_oflag  uint "output mode flags"
	c_cflag  uint "control mode flags"
	c_lflag  uint "local mode flags"
	c_line   byte "line discipline"
	c_cc[32] byte "control characters"
	c_ispeed uint "input speed"
	c_ospeed uint "output speed"
}

var oldtermios termios

func init_mouse() {
	fmt.Print("\x1B[?1000h")
}

func restore_mouse() {
	fmt.Print("\x1B[?1000l")
}

func init_terminal() {
	syscall.Syscall(syscall.SYS_IOCTL, 0, tcgets, uintptr(unsafe.Pointer(&oldtermios)))
	var newtermios = oldtermios
	newtermios.c_lflag &^= 010 // turn off echo
	newtermios.c_lflag &^= 02  // turn off canonical mode
	newtermios.c_cc[6] = 1     // set VMIN to 1
	newtermios.c_cc[5] = 0     // set VTIME to 0
	syscall.Syscall(syscall.SYS_IOCTL, 0, tcsetsf, uintptr(unsafe.Pointer(&newtermios)))
	
}

func restore_terminal() {
	restore_mouse()
	syscall.Syscall(syscall.SYS_IOCTL, 0, tcsetsf, uintptr(unsafe.Pointer(&oldtermios)))
}


func check_error(err os.Error) {
	if err != nil {
		fmt.Printf("Error: %s\n", err.String())
		panic(err.String())
	}
}
