package main

import "fmt"
import "log"
import "net"
import "os"
import "strconv"
import netstructs "./network_structs"
import "./minefield"

var rows, cols, mines int32
var player_conns[2] net.Conn
var field *minefield.Minefield
var player_field *minefield.Minefield
var mines_found_by_player[2] int

func main() {
	defer func() {
		if recover() != nil {
			fmt.Println("Exited.")
			os.Exit(1)
		}
	}()
	rows, cols, mines = 16, 16, 53
	handle_cmd_arguments()
	log.Printf("rows, cols, mines = %d %d %d\n", rows, cols, mines)
	wait_for_players()
	send_initial_packet_to_players()
	create_fields()
	game_loop()
	disconnect_players()
}

func game_loop() {
	var cur_player int32 = 0
	var mines_found int32 = 0
	for mines_found < mines {
		var f netstructs.Field
		f.Next_player = cur_player
		f.Scores[0] = int32(mines_found_by_player[0])
		f.Scores[1] = int32(mines_found_by_player[1])
		minefield.ToByteSlice(player_field, &f.Fields)
		netstructs.SendField(&f, player_conns[cur_player])
		netstructs.SendField(&f, player_conns[1-cur_player])
		move := netstructs.RecvMove(player_conns[cur_player])
		r, c := move.Row, move.Col
		log.Printf("Received move (%2d, %2d) from player %d.\n",
			r, c, cur_player)

		bounds_ok := !(r < 0 || r >= rows || c < 0 || c >= cols)
		if !bounds_ok || player_field.Fields[r][c] != '?' {
			log.Printf("Ignoring that move.")
			continue
		}

		if field.Fields[r][c] != 'x' {
			cur_player = 1 - cur_player
		} else {
			mines_found_by_player[cur_player] += 1
			mines_found += 1
		}
		reveal_fields(r, c)
	}

	var f netstructs.Field
	if mines_found_by_player[0] > mines_found_by_player[1] {
		f.Next_player = 0 - 2
	} else {
		f.Next_player = 1 - 2
	}
	minefield.ToByteSlice(player_field, &f.Fields)
	f.Scores[0] = int32(mines_found_by_player[0])
	f.Scores[1] = int32(mines_found_by_player[1])
	netstructs.SendField(&f, player_conns[0])
	netstructs.SendField(&f, player_conns[1])
}

func reveal_fields(r, c int32) {
	player_field.Fields[r][c] = field.Fields[r][c]
	if field.Fields[r][c] != 0 {
		return
	}
	for i := -1; i <= 1; i++ {
		for j := -1; j <= 1; j++ {
			rr := r + int32(i)
			cc := c + int32(j)
			row_ok := rr >= 0 && rr < rows
			col_ok := cc >= 0 && cc < cols
			if !row_ok || !col_ok {
				continue
			}
			if player_field.Fields[rr][cc] != '?' {
				continue
			}
			reveal_fields(rr, cc)
		}
	}
}

func create_fields() {
	log.Println("Generating minefield.")
	field = minefield.Create(rows, cols, mines)
	player_field = minefield.Create(rows, cols, mines)
	for r := 0; int32(r) < rows; r++ {
		for c := 0; int32(c) < cols; c++ {
			player_field.Fields[r][c] = '?'
		}
	}
}

func handle_cmd_arguments() {
	if len(os.Args) >= 2 && (os.Args[1] == "-h" || os.Args[1] == "--help") {
		fmt.Println("usage  : ", os.Args[0], " [rows] [cols] [mines]")
		fmt.Printf("default: %s %d %d %d\n",
			os.Args[0], rows, cols, mines)
		os.Exit(0)
	}

	if len(os.Args) >= 2 {
		i, _ := strconv.Atoi(os.Args[1])
		rows = int32(i)
	}
	if len(os.Args) >= 3 {
		i, _ := strconv.Atoi(os.Args[2])
		cols = int32(i)
	}
	if len(os.Args) >= 4 {
		i, _ := strconv.Atoi(os.Args[3])
		mines = int32(i)
	}

	if rows <= 0 || cols <= 0 || rows > 36 || cols > 36 || mines <= 0 || mines > rows*cols {
		fmt.Println("rows, cols and mines must be positive number less than or equal to 36")
		fmt.Println("and mines must be less or equal to rows*cols!")
		os.Exit(1)
	}
}

func wait_for_players() {
	log.Println("Creating listener.")
	listener, error := net.Listen("tcp", ":"+strconv.Itoa(netstructs.Port))
	check_error(error)

	log.Println("Waiting for player 0.")
	player_conns[0], error = listener.Accept()
	check_error(error)
	log.Println("Waiting for player 1.")
	player_conns[1], error = listener.Accept()
	check_error(error)
	listener.Close()
	log.Println("Closing listener.")
}

func disconnect_players() {
	log.Println("Disconnecting players.")
	for _, conn := range player_conns {
		conn.Close()
	}
}

func send_initial_packet_to_players() {
	for i := 0; i < 2; i++ {
		desc := netstructs.GameDescription{rows, cols, mines, int32(i)}
		netstructs.SendGameDescription(desc, player_conns[i])
	}
}

func check_error(err os.Error) {
	if err != nil {
		fmt.Printf("Error: %s\n", err.String())
		os.Exit(1)
	}
}
